#
# trace TCL script
#
# cse 457 autumn 1999 project 3
#
# douglas zongker
# dougz@cs.washington.edu
#

set g_idle -1

# unload_model is called to close a model.
proc unload_model {} {
    global g_idle    

    # destroy the output window
    after cancel $g_idle
    destroy .gl
}

# select_model is called to allow the user to pick a model to
# load.
proc select_model {} {
    # use the default file picker to get a DLL
    set fn [tk_getOpenFile -defaultextension "ray" -filetypes [list {{RAY} {.ray}}]]

    # if user hit cancel, return
    if {[string compare $fn ""]==0} return

    # unload the current model, if any

    unload_model

    # load the new model
    trace::getscene $fn

    wm title . "Trace - $fn"
}


# build_menubar is called, not surprisingly, to set up the main window's
# menu bar.
proc build_menubar {} {
    
    # create the root menu
    menu .mb -tearoff false

    # file menu
    menu .mb.file -tearoff false
    .mb.file add command -label "Load Trace File" -underline 0 -command select_model
    .mb.file add command -label "Save Image" -underline 0 -command trace::save
    .mb.file add command -label "Batch" -underline 0 -command trace::batch
    .mb.file add separator

    .mb.file add command -label "Quit" -underline 0 -command exit
    .mb add cascade -label "File" -underline 0 -menu .mb.file

}

proc draw {} {
    global g_width g_height g_idle
    trace::stoprender
    after cancel $g_idle
    destroy .gl
    soggy .gl -width $g_width -height $g_height -resize trace::resize -redraw trace::redraw
    pack .gl -in .right
    trace::render
    set g_idle [after idle continuous_redraw]
}

proc continuous_redraw {} {
    global g_idle
    .gl redraw
    set g_idle [after idle continuous_redraw]
}

###
### here is the main script -- not much to it, is there?
###
### this just sets up the main window; everything else
### is handled through callbacks.
###

global g_width g_height g_depth g_threshold

build_menubar
. configure -menu .mb
frame .c
frame .right

pack .c -side left
pack .right -expand true


if {[catch {load trace}]} {
wm withdraw .
tk_dialog .error "Error" "Can't find Trace.dll!" error 0 Quit
exit
}
if {[catch {load soggy}]} {
wm withdraw .
tk_dialog .error "Error" "Can't find soggy.dll!" error 0 Quit
exit
}

scale .width -label Width -from 32 -to 512 -variable g_width -showvalue true -orient horizontal -length 3i
pack .width -in .c
scale .height -label Height -from 32 -to 512 -variable g_height -showvalue true -orient horizontal -length 3i
pack .height -in .c    

button .renderbutton -text Render -command draw
pack .renderbutton -side bottom -in .c

button .stopbutton -text Stop -command trace::stoprender
pack .stopbutton -in .c



# if user gave a model name on the command line, try to load it.
if {[llength $argv]>0} {
    load_model [lindex $argv 0]
}

