#
# modeler TCL script
#
# cse 457 spring 1999 project 4
#
# douglas zongker
# dougz@cs.washington.edu
#

# substantial changes by Will Portnoy

# default "animation" menu settings:  animation off.
set g_timebase						-1
set g_currtime						0
set g_timerun						0
set g_running						0
set g_timebasediff					0
set g_usecurves						0
set g_majormode						"control"
set g_curve_editor_up				0
set g_currently_selected_curve_name	""

set x_division_interval				0

set y_division_interval				0

if {[catch {load modeler}]} {
	wm withdraw .
	tk_dialog .error "Error" "modeler.dll is not compiled.!" error 0 Quit
	exit
    }

proc change_timerun {{args}} {
    global g_timerun g_timebase g_currtime g_running
    global g_timebasediff
    global g_lastframe g_interval
    global g_fps

    if {$g_timerun && $g_running==0} {
	# starting the animation

	set g_timebase [expr [clock clicks]-$g_currtime*1000.0]
	set g_lastframe [clock clicks]
	set g_interval [expr 1000/$g_timerun]
	set g_fps $g_timerun
	set g_running 1
	
	timed_redraw
    }
}

# timed_redraw is called periodically when animation is on.
# it redraws the model, then schedules itself to run again.
proc timed_redraw {} {
    global g_timerun g_timebase g_currtime
    global g_maximum_animation_time g_usecurves g_controls
    global g_lastframe g_interval g_fps

    set g_currtime [expr ([clock clicks]-$g_timebase)/1000.0]
    while {$g_currtime > $g_maximum_animation_time} {
	set g_timebase [expr $g_timebase + $g_maximum_animation_time*1000]
	set g_currtime [expr $g_currtime - $g_maximum_animation_time]
    }

    redraw
    
    set now [clock clicks]
    if {[winfo exists .fps_value]} {
	if {$now == $g_lastframe} {
	    .fps_value config -text "inf"
	} else {
	    set g_fps [expr 0.95*$g_fps + 0.05*(1000.0/($now-$g_lastframe))]
	    .fps_value config -text [format "%.1f" $g_fps]
	}
    }

    if {$g_fps < $g_timerun} {
	incr g_interval -1
	if {$g_interval < 0} {set g_interval 0}
    } else {
	incr g_interval
    }

    if {$g_timerun} {
	after $g_interval timed_redraw
    } else {
	global g_running
	set g_running 0
    }
    
    set g_lastframe $now
}

# redraw invalidates the output window, forcing the model to
# redraw itself.  values for the control variables are taken
# from the curves if in curves mode.
proc redraw {} {
    global g_currtime g_usecurves g_controls g_curve_editor_up g_currently_selected_curve_name
	global minimum_time maximum_time time_interval
	global minimum_value maximum_value value_interval
	global shift_x shift_y

    if {$g_usecurves} {
	CurveEditor::updateCurveCurrentPositions $g_currtime g_controls
    }

	if {$g_curve_editor_up==1} {
		.curve_editor redraw

		.editor_axes delete axis_labels

		set stop_time_labels [expr $maximum_time - ($time_interval / 2.0)]

		for {set current_time_value $minimum_time} {$current_time_value < $stop_time_labels} {set current_time_value [expr $current_time_value + $time_interval]} {

			set start_x [expr 400.0 * $shift_x / ($maximum_time - $minimum_time)]
			#[expr 400.0 * $time_interval / ($maximum_time - $minimum_time)]

			set x	[expr 50 + int($start_x + (400.0 * (($current_time_value - $minimum_time ) / ($maximum_time  - ($minimum_time - $time_interval)))))]
			set y	420

			.editor_axes create text $x $y -text $current_time_value -tags [list axis_labels] -anchor center
		}

		set stop_y_labels [expr $maximum_value - ($value_interval / 2.0)]

		for {set current_y_value $minimum_value} {$current_y_value < $stop_y_labels} {set current_y_value [expr $current_y_value + $value_interval]} {

			set start_y	[expr 400.0 * $shift_y / ($maximum_value - $minimum_value)]
			#[expr 400.0 * $value_interval / ($maximum_value - $minimum_value)]

			set x	20
			set y	[expr 400 - int($start_y + (400.0 * (($current_y_value - $minimum_value) / ($maximum_value - ($minimum_value - $value_interval)))))]

			.editor_axes create text $x $y -text $current_y_value -tags [list axis_labels] -anchor center
		}

	}

    .gl redraw
}

# set_animation_time manually sets the animation time, and
# redraws the model at that time.
proc set_animation_time {time} {
    global g_currtime g_timebase
    set g_currtime $time
    set g_timebase [expr int([clock clicks]-$time*1000.0)]
    redraw
}

# reset_time changes the base time to the current time, which
# effectively resets the animation time to 0.0
proc reset_time {} {
    global g_timerun g_timebase g_currtime g_timebasediff
    set g_timebase [clock clicks]
    set g_timebasediff 0
    if {$g_timerun==0} {

		redraw
    }
}


proc write_ray_batch {} {
    global g_currtime g_maximum_animation_time
    global g_drawquality g_drawmode g_rayfilename g_timerun g_currtime

    set c [expr 1]

    while {$g_currtime < $g_maximum_animation_time} {
        # Advance the time

        set time [expr (0.1 + $g_currtime)]
        set_animation_time $time

        # Write the .ray file

        set fn "frame"
        append fn $c
        append fn ".ray"
        set g_rayfilename $fn
        set g_drawmode [expr 8|$g_drawquality]
        if {$g_timerun} {
	    set g_currtime [clock clicks]
        }

        .gl redraw
        update

        set_drawmode

        set c [expr $c + 1]
    }
}


# control_change is called whenever the value of a control
# widget changes.  It forces redraw of the model.
proc control_change {{args}} {

	redraw
}

proc toggle_show_animation_curve {item} {
	global g_currently_selected_curve_name

	CurveEditor::toggleShowAnimationCurve $item;

	 set g_currently_selected_curve_name [CurveEditor::getCurrentlySelectedCurveName]

	redraw
}

# load_model is called with the name of a model DLL to open.
proc load_model {dll} {
    global g_drawmode g_drawtype g_drawquality g_majormode

    destroy .gl
    
    # load the binary
    if {[catch [list load $dll]]} {
	tk_dialog .error "Error" "Failed to load model \"$dll\"!" error 0 Ok
	return
    }

    # call the model's setup function
    model::setup

	# create an output window and attach the model's redraw
    # and resize functions to it.

    soggy .gl -width 400 -height 400 -resize model::resize -redraw model::redraw

    pack .gl -in .model_display -side top -padx 2m -pady 2m

	set g_majormode curve;

	change_majormode

	redraw
}

proc set_curve_type {item} {
	global .curve_editor

	CurveEditor::setCurrentlySelectedCurveType $item;

	redraw
}

proc set_time_bar {x} {

	set time [CurveEditor::getTime $x]

	set_animation_time $time
}

proc change_animation_length {} {
	global g_maximum_animation_time

	CurveEditor::changeAnimationLength $g_maximum_animation_time

	redraw
}

proc scale_curves {} {
	global g_scaling_factor g_maximum_animation_time

	CurveEditor::rescaleTime $g_scaling_factor

	set g_maximum_animation_time [expr $g_maximum_animation_time * $g_scaling_factor]

	change_animation_length

	redraw
}


proc try_pop_zoom {} {

	CurveEditor::tryPopZoom

	redraw
}

proc toggle_wrap_control_points {} {

	CurveEditor::toggleWrapControlPoints

	redraw
}

proc change_minimum_y {} {
	global new_y_minimum

	CurveEditor::changeMinimumY $new_y_minimum

	redraw
}

proc change_maximum_y {} {
	global new_y_maximum

	CurveEditor::changeMaximumY $new_y_maximum

	redraw
}

proc setup_curve_editor {} {
	global minimum_value maximum_value minimum_time maximum_time 
	global g_curve_editor_up g_currently_selected_curve_name g_maximum_animation_time g_scaling_factor
	global new_y_minimum new_y_maximum

	canvas .editor_axes -width 450 -height 450

	pack .editor_axes -in .editor_and_axes


	soggy .curve_editor -width 400 -height 400 -resize CurveEditor::resize -redraw CurveEditor::redraw

	set g_curve_editor_up	1

	place .curve_editor -in .editor_and_axes -x 50 -y 0
	
	.curve_editor redraw

	bind .curve_editor <ButtonPress-1>			{CurveEditor::mouseButtonDown %x %y; set g_currently_selected_curve_name [CurveEditor::getCurrentlySelectedCurveName]; redraw}
	bind .curve_editor <Button1-Motion>			{CurveEditor::mouseDrag %x %y; set g_currently_selected_curve_name [CurveEditor::getCurrentlySelectedCurveName]; redraw}
	bind .curve_editor <ButtonRelease-1>		{CurveEditor::mouseButtonUp %x %y; set g_currently_selected_curve_name [CurveEditor::getCurrentlySelectedCurveName]; redraw}

	bind .curve_editor <Control-1>				{CurveEditor::controlClick %x %y; set g_currently_selected_curve_name [CurveEditor::getCurrentlySelectedCurveName]; redraw}
	bind .curve_editor <ButtonPress-2>			{CurveEditor::altClick %x %y; set g_currently_selected_curve_name [CurveEditor::getCurrentlySelectedCurveName]; redraw}

	bind .curve_editor <Shift-ButtonPress-1>	{CurveEditor::shiftMouseButtonDown %x %y; set g_currently_selected_curve_name [CurveEditor::getCurrentlySelectedCurveName]; redraw}
	bind .curve_editor <Shift-Button1-Motion>	{CurveEditor::shiftMouseDrag %x %y; set g_currently_selected_curve_name [CurveEditor::getCurrentlySelectedCurveName]; redraw}
	bind .curve_editor <Shift-ButtonRelease-1>	{CurveEditor::shiftMouseButtonUp   %x %y; set g_currently_selected_curve_name [CurveEditor::getCurrentlySelectedCurveName]; redraw}

	menu .curve_editor.menu -tearoff false

	.curve_editor.menu add cascade -label "Curve types" -menu .curve_editor.menu.types
	menu .curve_editor.menu.types -tearoff false

	foreach item [CurveEditor::getCurveTypes] {
		.curve_editor.menu.types add command -label $item -command [list set_curve_type $item]
	}

	bind .curve_editor <Shift-ButtonPress-2> [list set_time_bar %x]

	bind .curve_editor <Shift-Button2-Motion> [list set_time_bar %x]

	bind .curve_editor <Shift-ButtonRelease-2> [list set_time_bar %x]


	bind .curve_editor <ButtonPress-3> {tk_popup .curve_editor.menu %X %Y; set g_currently_selected_curve_name [CurveEditor::getCurrentlySelectedCurveName]}


	label .currently_selected_curve_label -text "Currently Selected Curve: "
	label .currently_selected_curve_value -width 15 -textvariable g_currently_selected_curve_name

	pack .currently_selected_curve_label -in .currently_selected_curve_frame -side left
	pack .currently_selected_curve_value -in .currently_selected_curve_frame -side left



	label .currently_selected_curve_number_of_control_points_label -text "Number of Control Points: "
	label .currently_selected_curve_number_of_control_points_value -width 15 -textvariable number_of_control_points

	pack .currently_selected_curve_number_of_control_points_label -in .currently_selected_curve_number_of_control_points_frame -side left
	pack .currently_selected_curve_number_of_control_points_value -in .currently_selected_curve_number_of_control_points_frame -side left


	label .currently_selected_curve_number_of_curve_segments_label -text "Number of Curve Segments: "
	label .currently_selected_curve_number_of_curve_segments_value -width 15 -textvariable number_of_curve_segments

	pack .currently_selected_curve_number_of_curve_segments_label -in .currently_selected_curve_number_of_curve_segments_frame -side left
	pack .currently_selected_curve_number_of_curve_segments_value -in .currently_selected_curve_number_of_curve_segments_frame -side left




	label .fps_label -text "FPS: "
	label .fps_value -width 5 -text ""

	pack .fps_label -in .fps_frame -side left
	pack .fps_value -in .fps_frame -side left



	button .undo_zoom -text "Undo Zoom" -command try_pop_zoom

	pack .undo_zoom -in .editor_controls_left -side top


	checkbutton .wrap_curves -text "Wrap Curves" -command toggle_wrap_control_points

	pack .wrap_curves -in .editor_controls_left -side top


	label .maximum_animation_time_label -text "Maximum Animation Time" -width 20
	entry .maximum_animation_time_entry -width 5 -relief sunken -bd 2 -textvariable g_maximum_animation_time
	button .change_animation_length_button -text "Apply" -command change_animation_length

	pack .change_animation_length_button -in .maximum_animation_time_frame -side right
	pack .maximum_animation_time_entry -in .maximum_animation_time_frame -side right
	pack .maximum_animation_time_label -in .maximum_animation_time_frame -side right



	label .scaling_label -text "Scaling Factor" -width 20
	entry .scaling_entry -width 5 -relief sunken -bd 2 -textvariable g_scaling_factor
	button .scaling_length_button -text "Apply" -command scale_curves

	pack .scaling_length_button -in .scaling_frame -side right
	pack .scaling_entry -in .scaling_frame -side right
	pack .scaling_label -in .scaling_frame -side right



	label .change_minimum_y_label -text "Minimum Value" -width 20
	entry .change_minimum_y_entry -width 5 -relief sunken -bd 2 -textvariable new_y_minimum
	button .change_minimum_y_button -text "Apply" -command change_minimum_y

	pack .change_minimum_y_button -in .change_minimum_y_frame -side right
	pack .change_minimum_y_entry -in .change_minimum_y_frame -side right
	pack .change_minimum_y_label -in .change_minimum_y_frame -side right

	label .change_maximum_y_label -text "Maximum Value" -width 20
	entry .change_maximum_y_entry -width 5 -relief sunken -bd 2 -textvariable new_y_maximum
	button .change_maximum_y_button -text "Apply" -command change_maximum_y

	pack .change_maximum_y_button -in .change_maximum_y_frame -side right
	pack .change_maximum_y_entry -in .change_maximum_y_frame -side right
	pack .change_maximum_y_label -in .change_maximum_y_frame -side right

	.curve_editor.menu add cascade -label "Animation variables" -menu .curve_editor.menu.variables
	menu .curve_editor.menu.variables -tearoff false

	foreach item [CurveEditor::getCurveNames] {
		.curve_editor.menu.variables add checkbutton -label $item -command [list toggle_show_animation_curve $item]
	}
}

proc cleanup_curve_editor {} {
	global g_curve_editor_up

	destroy .curve_editor

	destroy .editor_axes

	set g_curve_editor_up	0

	destroy .curve_editor.menu.variables
	destroy .curve_editor.menu.types
	destroy .curve_editor.menu

	destroy .currently_selected_curve_label
	destroy .currently_selected_curve_value

	destroy .fps_label
	destroy .fps_value

	destroy .currently_selected_curve_number_of_control_points_label
	destroy .currently_selected_curve_number_of_control_points_value

	destroy .currently_selected_curve_number_of_curve_segments_label
	destroy .currently_selected_curve_number_of_curve_segments_value

	destroy .undo_zoom

	destroy .wrap_curves

	destroy .maximum_animation_time_label
	destroy .maximum_animation_time_entry
	destroy .change_animation_length_button

	destroy .scaling_label
	destroy .scaling_entry
	destroy .scaling_length_button

	destroy .change_maximum_y_button
	destroy .change_maximum_y_entry
	destroy .change_maximum_y_label

	destroy .change_minimum_y_button
	destroy .change_minimum_y_entry
	destroy .change_minimum_y_label

}

# unload_model is called to close a model.
proc unload_model {} {
    # stop the animation
    global g_timerun
    set g_timerun 0

	cleanup_curve_editor

	set g_curve_editor_up 0

	destroy .change_animation_length_button
	destroy .undo_zoom
	destroy .wrap_curves 

    # destroy the output window

    destroy .gl

    # reset the modeler (which deletes all control widgets)
    model::reset
}

# select_model is called to allow the user to pick a model to
# load.
proc select_model {} {
    # use the default file picker to get a DLL
    set fn [tk_getOpenFile -defaultextension "dll" -filetypes [list {{DLLs} {.dll}}]]

    # if user hit cancel, return
    if {[string compare $fn ""]==0} return

    # unload the current model, if any
    close_model

    # load the new model
    load_model $fn

    redraw
}

# close_model is called to remove the current model (either by
# user request, or when a new model is about to be loaded)
proc close_model {} {
    global g_majormode
    set g_majormode "control"
    change_majormode
    CurveEditor::clearCurves

    # unload the model
    unload_model
}

# write_raytrace lets user select a file for output, then writes the
# current model as a raytracer input file.
proc write_raytrace {} {
    global g_drawquality g_drawmode g_rayfilename g_timerun g_currtime
    set fn [tk_getSaveFile -defaultextension "ray" -filetypes [list {{Ray tracer input files} {.ray}}]]
    if {[string compare $fn ""]==0} return

    set g_rayfilename $fn
    set g_drawmode [expr 8|$g_drawquality]
    if {$g_timerun} {
	set g_currtime [clock clicks]
    }

    .gl redraw
    update

    set_drawmode
}

# save_animation lets user select a file for output, then writes the
# current animation curves to the file.
proc save_animation {} {
    set fn [tk_getSaveFile -defaultextension "anm" -filetypes [list {{Animation files} {.anm}}]]
    if {[string compare $fn ""]==0} return

    if {[catch {open $fn w} fd]} {
	tk_dialog .error Error "couldn't open $fn." error 0 ok
	return
    }

    CurveEditor::saveCurves $fn
    close $fd

	redraw
}

# load_animation loads animation curves from a user-selected file.
proc load_animation {} {
    set fn [tk_getOpenFile -defaultextension "anm" -filetypes [list {{Animation files} {.anm}}]]
    if {[string compare $fn ""]==0} return

    if {[catch {open $fn} fd]} {
	tk_dialog .error Error "couldn't open $fn." error 0 ok
	return
    }

    CurveEditor::loadCurves $fn
    close $fd

	redraw
}

# set_drawmode is called when something on the view menu is
# selected.  It updates the g_drawmode variable with the
# new bit flags.
proc set_drawmode {{args}} {
    global g_drawmode g_drawtype g_drawquality
    
    # the drawing MODE is the drawing TYPE (normal, wireframe, flatshade)
    # flag OR'd together with the QUALITY (high, medium, low)
    set g_drawmode [expr $g_drawtype|$g_drawquality]

    redraw
}

# change_majormode is called to switch between "control" mode and "curve"
# mode.  It handles swapping widgets in the main window.
proc change_majormode {} {
    global g_majormode g_usecurves

    switch $g_majormode {
	control {
	    if {[winfo exists .cv]} {
		destroy .cv
	    }
	    if {![winfo exists .c]} {
		frame .c -width 4i
		pack .c
	    }
	    catch model::rebuild
	    set g_usecurves 0
		cleanup_curve_editor
	    .mb.file entryconfigure 3 -state normal
	    .mb.file entryconfigure 4 -state normal
	}
	curve {
	    if {[winfo exists .c]} {
		destroy .c
	    }
	    frame .cv
	    pack .cv -expand 1 -fill both
	    set g_usecurves 1
	    .mb.file entryconfigure 3 -state normal
	    .mb.file entryconfigure 4 -state normal
		setup_curve_editor
		redraw
	}
    }
}

# build_menubar is called, not surprisingly, to set up the main window's
# menu bar.
proc build_menubar {} {
    global g_drawtype g_drawquality g_timerun
    
    # create the root menu
    menu .mb -tearoff false

    # file menu
    menu .mb.file -tearoff false
    .mb.file add command -label "Open model..." -underline 0 -command select_model
    .mb.file add command -label "Close model" -underline 0 -command close_model
    .mb.file add separator
    .mb.file add command -label "Load animation..." -underline 0 -command load_animation
    .mb.file add command -label "Save animation..." -underline 0 -command save_animation
    .mb.file add separator
    .mb.file add command -label "Write raytrace file..." -underline 7 -command write_raytrace
    .mb.file add command -label "Write raytrace batch" -underline 0 -command write_ray_batch
    .mb.file add separator
    .mb.file add command -label "Quit" -underline 0 -command exit
    .mb add cascade -label "File" -underline 0 -menu .mb.file

    # edit menu
    menu .mb.edit -tearoff false
    .mb.edit add command -label "Redraw" -underline 0 -command { .gl redraw }
    .mb.edit add separator
    .mb.edit add radiobutton -label "Control mode" -underline 0 -variable g_majormode -value control -command change_majormode
    .mb.edit add radiobutton -label "Curve mode" -underline 0 -variable g_majormode -value curve -command change_majormode
    .mb add cascade -label "Edit" -underline 0 -menu .mb.edit

    # view menu
    menu .mb.view -tearoff false
    .mb.view add radiobutton -label "Normal" -underline 0 -variable g_drawtype -value 1
    .mb.view add radiobutton -label "Wireframe" -underline 0 -variable g_drawtype -value 2
    .mb.view add radiobutton -label "Flat shaded" -underline 0 -variable g_drawtype -value 4
    .mb.view add separator
    .mb.view add radiobutton -label "High quality" -underline 0 -variable g_drawquality -value 256
    .mb.view add radiobutton -label "Medium quality" -underline 0 -variable g_drawquality -value 512
    .mb.view add radiobutton -label "Low quality" -underline 0 -variable g_drawquality -value 1024
    .mb add cascade -label "View" -underline 0 -menu .mb.view

    # default "view" menu settings: normal mode, high quality
    set g_drawtype    1      
    set g_drawquality 1024    

    # whenever a view menu item is changed, call the set_drawmode function
    trace variable g_drawtype w set_drawmode
    trace variable g_drawquality w set_drawmode

    # also, call it now, with the default values 
    set_drawmode

    # animation menu 
    menu .mb.animate -tearoff false
    .mb.animate add radiobutton -label "Animation off" -variable g_timerun -value 0 -underline 0
    .mb.animate add radiobutton -label "5 frames/second" -variable g_timerun -value 5 -underline 0
    .mb.animate add radiobutton -label "10 frames/second" -variable g_timerun -value 10 -underline 0
    .mb.animate add radiobutton -label "20 frames/second" -variable g_timerun -value 20 -underline 0
    .mb.animate add radiobutton -label "30 frames/second" -variable g_timerun -value 30 -underline 0
    .mb.animate add separator
    .mb.animate add command -label "Reset time" -underline 0 -command reset_time
    .mb.animate add separator
    .mb.animate add checkbutton -label "Use curves" -variable g_usecurves -underline 4
    .mb add cascade -label "Animate" -underline 0 -menu .mb.animate

    trace variable g_timerun w change_timerun
}

proc setup_frames {} {
	frame .c
	frame .right
	frame .editor_stuff
	frame .model_display

	pack .c -side left

	pack .right -side right

	pack .editor_stuff -in .right -side left

	pack .model_display -in .right -side right

	frame .editor_and_axes;

	pack .editor_and_axes -in .editor_stuff

	frame .editor_controls

	pack .editor_controls -in .editor_stuff -side bottom

	frame .editor_controls_left

	pack .editor_controls_left -in .editor_controls -side left

	frame .editor_controls_right

	pack .editor_controls_right -in .editor_controls -side right



	frame .currently_selected_curve_frame

	pack .currently_selected_curve_frame -in .editor_controls_left -side top

	frame .currently_selected_curve_number_of_control_points_frame

	pack .currently_selected_curve_number_of_control_points_frame -in .editor_controls_left -side top

	frame .currently_selected_curve_number_of_curve_segments_frame

	pack .currently_selected_curve_number_of_curve_segments_frame -in .editor_controls_left -side top


	frame .fps_frame

	pack .fps_frame -in .editor_controls_left -side top


	frame .maximum_animation_time_frame

	pack .maximum_animation_time_frame -in .editor_controls_right -side top

	frame .scaling_frame

	pack .scaling_frame -in .editor_controls_right -side top

	frame .change_minimum_y_frame

	pack .change_minimum_y_frame -in .editor_controls_right -side top

	frame .change_maximum_y_frame

	pack .change_maximum_y_frame -in .editor_controls_right -side top
}

###
### here is the main script -- not much to it, is there?
###
### this just sets up the main window; everything else
### is handled through callbacks.
###

setup_frames

load "soggy"

soggy .gl -width 400 -height 400

pack .gl -in .model_display

load "Debug\\CurveEditor"

set g_scaling_factor 1.0;

set g_maximum_animation_time 20;

set minimum_time				0
set maximum_time				0
set time_interval				0
set minimum_value				0
set maximum_value				0
set value_interval				0

set shift_x						0
set shift_y						0
set number_of_control_points	0
set number_of_curve_segments	0

# create the menu and put it in the window
build_menubar
. configure -menu .mb

# initialize the main window in "control" mode
change_majormode

# if user gave a model name on the command line, try to load it.
if {[llength $argv]>0} {
    load_model [lindex $argv 0]
}

# if exists

# load_model cylinder.dll